% Tracking of objects in a low-resolution grey-scale image
% Made by Sveinung Pettersen Rsok
% 09.06.2015

% Full program with all functions

%saveas(gcf,'mutual_nodes_461_2_result','png')
clear all;

                    %% Initiations for testing
min_dist_mode = 0;
counter = 0;
sim_length = 0;
object_1 = 0;
object_2 = 0;
validation_filter_object_1 = 0;
validation_filter_object_2 = 0;

% For Kalman filter
dT = .02;
A = [1 dT; 0 1] ; % state transition matrix
H = [1 0]; % observation matrix
I = [1 0;0 1]; % identity matrix
C = 1;
dev_xhat  = .01;
dev_xdoublehat = .1;
Q = [dev_xhat.^2 0;0 dev_xdoublehat.^2];

M_x1 = [0 0;0 0];
M_y1 = [0 0;0 0];
M_x2 = [0 0;0 0];
M_y2 = [0 0;0 0];
x1_updated = 0;  %  filtered position
y1_updated = 0;  %  filtered position
x2_updated = 0;  %  filtered position
y2_updated = 0;  %  filtered position
x1initial_values = [0; 0]; % Initial [position; velocity]
y1initial_values = [0; 0]; % Initial [position; velocity]
x2initial_values = [0; 0]; % Initial [position; velocity]
y2initial_values = [0; 0]; % Initial [position; velocity]
x1est_state = x1initial_values;
y1est_state = y1initial_values;
x2est_state = x2initial_values;
y2est_state = y2initial_values;
Kalman_object_1 = 0;
Kalman_object_2 = 0;

% 37x22 nodes, cutout from Atmel touch display
indata2=csvread('finger_stylus.csv',1,8); % 1-500 0 single finger etter dette?
indata3=csvread('finger_lines.csv',1,8); % 18-111
indata4=csvread('pinch_split.csv',1,8); % 24-->694 Vis 550:694
indata5=csvread('steady_finger.csv',1,8); % 15-200, 213-350, 370-490fail, 520-615,630-790
indata6=csvread('stylus_lines.csv',1,8); % 20-212
indata7=csvread('diagonal.csv',1,8); % Finger:23-130, 147-253 Stylus: 303-461, 491-613
indata8=csvread('rotate.csv',1,8); % 1-85, 105-215, 245-436
indata9=csvread('approach-depart.csv',1,8); % 5-145, 185-277, 310-415, 

%% Show tracking, frame by frame
plot_step_by_step = 1;
%% Show history
plot_history = 1;

% Length of data sets (i)
for i = 24:694
    for j=1:37
        original(j,1:22)=indata4(i,(j-1)*22+[2:23]);
    end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                        %% MY TRACKING PROGRAM
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                        %% WORKER UNIT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is the first part of the program, which does stand-alone searches of
% each dataset. It recieves parameters from the other part, the smart unit,
% and performs the search based on these parameters.

% 3 main functions: Find 1, find 2, find find 2 close objects
                    
track_img = original;
matrix_size = size(track_img);
sim_length = sim_length +1;

% Remove negative values
for k = 1:matrix_size(1)
    for l = 1:matrix_size(2)
        if track_img(k,l) < 0
            track_img(k,l) = 0;
        end
    end
end


%% Find 1, input(track_img), output(x0y0)
%track = find_1(track_img);

%% Find 2, input(track_img), output(x0y0/x1y1)
if min_dist_mode == 0
    tracks = find_2(track_img);
    track_1 = [tracks(1) tracks(2)];
    track_2 = [tracks(3) tracks(4)];

%% Fine positioning, below minimum distance, either LS or CoM
elseif min_dist_mode == 1
    tracks_close = find_2_close(track_img, prev_obj1_node(1), prev_obj1_node(2), prev_obj2_node(1), prev_obj2_node(2));
    track_1(1) = tracks_close(1); track_1(2) = tracks_close(2);
    track_2(1) = tracks_close(3); track_2(2) = tracks_close(4);
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                        %% SMART UNIT %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Labelling
% Validation
% Check minimum distance
% Log previous estimated position and max_node for up to two objects
% Use Kalman filter

%% (1) Labelling - Connect to closest dot
if track_1(1) > 0 && sim_length > 1
    track_objects = connect_to_closest( track_1(1) , track_1(2) , track_2(1) , track_2(2) , prev_track1_x(sim_length-1) , prev_track1_y(sim_length-1) , prev_track2_x(sim_length-1) , prev_track2_y(sim_length-1) ) ;
    track_1(1) = track_objects(1);
    track_1(2) = track_objects(2);
    track_2(1) = track_objects(3);
    track_2(2) = track_objects(4);
end

if track_1(1) == 0
    object_1 = 0;
else
    object_1 = 1;
end

if track_2(1) == 0
    object_2 = 0;
else
    object_2 = 1;
end

%% (2) Validation filter

% Track 1
if object_1 == 0
    validation_filter_object_1 = 0;
    x1initial_values = [0; 0]; % Initial [position; velocity]
    y1initial_values = [0; 0]; % Initial [position; velocity]
    Kalman_object_1 = 0;
elseif object_1 == 1 && validation_filter_object_1 < 3
    validation_filter_object_1 = validation_filter_object_1 + 1;
    object_1 = 0;
end

% Track 2
if object_2 == 0
    validation_filter_object_2 = 0;
    x2initial_values = [0; 0]; % Initial [position; velocity]
    y2initial_values = [0; 0]; % Initial [position; velocity]
    Kalman_object_2 = 0;
elseif object_2 == 1 && validation_filter_object_2 < 3
    validation_filter_object_2 = validation_filter_object_2 + 1;
    object_2 = 0;
end


%% (3)  Check minimum distance (should be around 4.6, ref report)
if min_dist_mode == 1 && sqrt((track_1(1)-track_2(1)).^2 + (track_1(2)-track_2(2)).^2) > 6
    min_dist_mode = 0;
elseif object_1 == 1 && object_2 == 1 && sqrt((track_1(1)-track_2(1)).^2 + (track_1(2)-track_2(2)).^2) < 5;
    min_dist_mode = 1;
else
    min_dist_mode = 0;
end
    

%% (4)  Log current position
prev_track1_x(sim_length) = track_1(1);
prev_track1_y(sim_length) = track_1(2);
prev_track2_x(sim_length) = track_2(1);
prev_track2_y(sim_length) = track_2(2);

% Make sure they do not have the same main node
if object_1 == 1 & object_2 == 1 && round(track_2(1))== round(track_1(1)) && round(track_2(2))==round(track_1(2))
    prev_obj1_node = [round(track_1(1)) round(track_1(2))];
    prev_obj2_node = prev_obj2_node; 
else
    prev_obj1_node = [round(track_1(1)) round(track_1(2))];
    prev_obj2_node = [round(track_2(1)) round(track_2(2))];
end


%% (5)  Kalman filter
% Object 1
if object_1 == 1 && Kalman_object_1 == 0
    est_state_x1 = [track_1(1);0];
    est_state_y1 = [track_1(2);0];
    Kalman_object_1 = 1;
    x1_updated = track_1(1);
    y1_updated = track_1(2);
elseif object_1 == 1
% x1
    est_state_x1 = A * est_state_x1;
    M_x1 = A * M_x1 * A' + Q;
    K = M_x1*H'*inv(H*M_x1*H'+C);
    est_state_x1 = est_state_x1 + K * (track_1(1) - H * est_state_x1);
    M_x1 =  (I-K*H)*M_x1;
    x1_updated = est_state_x1(1);
% y1
    est_state_y1 = A * est_state_y1;
    M_y1 = A * M_x1 * A' + Q;
    K = M_y1*H'*inv(H*M_y1*H'+C);
    est_state_y1 = est_state_y1 + K * (track_1(2) - H * est_state_y1);
    M_y1 =  (I-K*H)*M_y1;
    y1_updated = est_state_y1(1);
end

% Object 2
if object_2 == 1 && Kalman_object_2 == 0
    est_state_x2 = [track_2(1);0];
    est_state_y2 = [track_2(2);0];
    Kalman_object_2 = 1;
    x2_updated = track_2(1);
    y2_updated = track_2(2);
elseif object_2 == 1
% x2
    est_state_x2 = A * est_state_x2;
    M_x2 = A * M_x2 * A' + Q;
    K = M_x2*H'*inv(H*M_x2*H'+C);
    est_state_x2 = est_state_x2 + K * (track_2(1) - H * est_state_x2);
    M_x2 =  (I-K*H)*M_x2;
    x2_updated = est_state_x2(1);
% y2
    est_state_y2 = A * est_state_y2;
    M_y2 = A * M_y2 * A' + Q;
    K = M_y2*H'*inv(H*M_y2*H'+C);
    est_state_y2 = est_state_y2 + K * (track_2(2) - H * est_state_y2);
    M_y2 =  (I-K*H)*M_y2;
    y2_updated = est_state_y2(1);
end

% Log Kalman positions for graph
x1_updated_history(sim_length) = x1_updated;
y1_updated_history(sim_length) = y1_updated;
x2_updated_history(sim_length) = x2_updated;
y2_updated_history(sim_length) = y2_updated;

                            %% PLOT %%
% Plot active nodes + estimated centers on display
track_img_result = zeros(matrix_size(1), matrix_size(2));
act_nodes = 0;
if object_1 == 1
act_nodes = size_estimation(track_img, round(track_1(1)), round(track_1(2)));
track_img_result(act_nodes(3):act_nodes(4),act_nodes(1):act_nodes(2))= track_img(act_nodes(3):act_nodes(4),act_nodes(1):act_nodes(2)) ;
end
if object_2 == 1
act_nodes = size_estimation(track_img, round(track_2(1)), round(track_2(2)));
track_img_result(act_nodes(3):act_nodes(4),act_nodes(1):act_nodes(2)) = track_img(act_nodes(3):act_nodes(4),act_nodes(1):act_nodes(2));
end

% Only confirmed objects are visible on the middle and right screen
if plot_step_by_step == 1
figure(1);
show_tracking_final(track_img, track_img_result, object_1, track_1(1), track_1(2), object_2, track_2(1), track_2(2), x1_updated, y1_updated, x2_updated, y2_updated);
pause(.1);
end

end

if plot_history == 1
show_tracking_history(prev_track1_x, prev_track1_y, prev_track2_x, prev_track2_y, x1_updated_history, y1_updated_history, x2_updated_history, y2_updated_history, sim_length, matrix_size(2), matrix_size(1), 0);
end

